/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "native.pmh"

namespace fe
{

#if FE_2DGL==FE_2D_X_GFX

#define FE_X_EVENTMASK	ExposureMask		| \
						EnterWindowMask		| \
						LeaveWindowMask		| \
						KeyPressMask		| \
						KeyReleaseMask		| \
						ButtonPressMask		| \
						ButtonReleaseMask	| \
						StructureNotifyMask

/*** alternates:    PointerMotionMask/ButtonMotionMask
					StructureNotifyMask/SubstructureNotifyMask */


Result NativeWindow::nativeInitialize(void)
{
	FEASSERT(m_spEventContextI.isValid());
	sp<NativeEventContext> spNativeEventContext=m_spEventContextI;
	FEASSERT(spNativeEventContext.isValid());
	Display *pDisplay=spNativeEventContext->displayHandle();

	m_spEventContextI->threadLock();

	//* NOTE try to fit into 80% of one monitor

	Screen* pScreen=DefaultScreenOfDisplay(pDisplay);

//	const I32 screenWidth=XWidthOfScreen(pScreen);
	const I32 screenHeight=XHeightOfScreen(pScreen);

	const int windowX=0.1*screenHeight;
	const int windowY=0.1*screenHeight;
	const int windowHeight=0.8*screenHeight;
	const int windowWidth=1.2*windowHeight;
	setPosition(windowX,windowY);
	setSize(windowWidth,windowHeight);

	m_spEventContextI->threadUnlock();
	return e_ok;
}

Result NativeWindow::nativeStartup(void)
{
	//* in case something goes wrong
	FE_WINDOW_HANDLE default_handle={ NULL, NULL, 0 };
	setWindowHandle(default_handle);

	FEASSERT(m_spEventContextI.isValid());
	sp<NativeEventContext> spNativeEventContext=m_spEventContextI;
	FEASSERT(spNativeEventContext.isValid());
	Display *pDisplay=spNativeEventContext->displayHandle();

	m_spEventContextI->threadLock();
	const int screenIndex=DefaultScreen(pDisplay);

	Window parent=RootWindow(pDisplay,screenIndex);
	if(!parent)
	{
		feLog("NativeWindow::nativeStartup"
				" RootWindow(pDisplay=%p,screenIndex=%p)=NULL\n",
				pDisplay,screenIndex);
		return e_undefinedFailure;
	}

	//* NOTE presumably powers of 2, up to 64
	const I32 multisample=
			registry()->master()->catalog()->catalog<I32>("hint:multisample");
	I32 samples=(multisample>0)? pow(2,I32(log2(multisample))): 0;

	int m=0;
	int attributes[32];
	attributes[m++]=GLX_RGBA;
	attributes[m++]=GLX_RED_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_GREEN_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_BLUE_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_DEPTH_SIZE;
	attributes[m++]=1;
	attributes[m++]=GLX_DOUBLEBUFFER;

	//* MSAA
	if(samples)
	{
		attributes[m++]=GLX_SAMPLE_BUFFERS;
		attributes[m++]=1;
		attributes[m++]=GLX_SAMPLES;
		attributes[m++]=samples;
	}

	attributes[m]=None;

	XVisualInfo *pVisualInfo=glXChooseVisual(pDisplay,screenIndex,attributes);
	while(!pVisualInfo && samples>0)
	{
		samples=samples>>1;

		if(samples)
		{
			attributes[m-1]=samples;
		}
		else
		{
			m-=4;
			attributes[m]=None;
		}

		pVisualInfo=glXChooseVisual(pDisplay,screenIndex,attributes);
		if(pVisualInfo && !samples)
		{
			feLog("NativeWindow::nativeStartup multisampling not available\n");
		}
	}
	if(pVisualInfo && samples && samples!=multisample)
	{
		feLog("NativeWindow::nativeStartup"
				" using multisample %d instead of %d\n",samples,multisample);
	}
	if(!pVisualInfo)
	{
		feLog("NativeWindow::nativeStartup "
					"compatible visual not available through GLX.\n"
					"  Using default visual.\n");

		int nitems;
		XVisualInfo vinfo;
		vinfo.depth=DefaultDepth(pDisplay,screenIndex);
		vinfo.visual=DefaultVisual(pDisplay,screenIndex);
		pVisualInfo=XGetVisualInfo(pDisplay,VisualNoMask,&vinfo,&nitems);
	}

	unsigned long value_mask=0;
	XSetWindowAttributes cwa;

	cwa.border_pixel=0;
	cwa.colormap=XCreateColormap(pDisplay,parent,pVisualInfo->visual,
			AllocNone);

	value_mask=CWColormap|CWBorderPixel;

	//* create the window
	Window win=XCreateWindow(pDisplay,parent,
							m_box[0],m_box[1],			// position
							width(m_box),height(m_box),	// size
							0,							// border width
							pVisualInfo->depth,
							InputOutput,
							pVisualInfo->visual,
							value_mask,&cwa);

	FE_WINDOW_HANDLE handle={pDisplay, pVisualInfo, win};
	setWindowHandle(handle);

	setPointerMotion(spNativeEventContext->pointerMotion());

	//* set client name
	char hostname[128];
	int fail=gethostname(hostname,128);
	FEASSERT(!fail);
	XTextProperty clientName;
	char *stringlist[1];
	stringlist[0]=hostname;
	if(!XStringListToTextProperty(stringlist,1,&clientName))
	{
		feX("Could not allocate space for client name");
	}
	XSetWMClientMachine(pDisplay,win,&clientName);
	XFree((void *)clientName.value);

	//* set pid
	pid_t pid=getpid();
	Atom atom = XInternAtom(pDisplay, "_NET_WM_PID", FALSE);
	FEASSERT(atom!=None);
	XChangeProperty(pDisplay,win,atom,XA_CARDINAL,32,
			PropModeReplace,(unsigned char *)(&pid), 1);

	//* create GLX context
	GLXContext shareList=NULL;
	Bool direct=True;
	m_glxContext=glXCreateContext(pDisplay,pVisualInfo,shareList,direct);
	if(m_glxContext==NULL)
		feX("NativeWindow::nativeStartup","glXCreateContext failed\n");
	makeCurrent();

	spNativeEventContext->notifyCreate(m_windowHandle);

	//* force window to update now
	XFlush(pDisplay);
//	XSync(pDisplay,FALSE);

#ifdef FE_USE_OPENCL
	cl_uint platformIdCount=0;
	clGetPlatformIDs(0,NULL,&platformIdCount);
	feLog("platformIdCount %d\n",platformIdCount);

	if(platformIdCount>0)
	{
		std::vector<cl_platform_id> platformIdArray(platformIdCount);
		clGetPlatformIDs(platformIdCount,platformIdArray.data(),NULL);

		cl_platform_id clPlatformId=NULL;
		BWORD useGLX=FALSE;

		for(I32 platformIdIndex=0;platformIdIndex<platformIdCount;
				platformIdIndex++)
		{
			size_t name_size=0;
			clGetPlatformInfo(platformIdArray[platformIdIndex],
					CL_PLATFORM_NAME,0,NULL,&name_size);

			char* platform_name=new char[name_size+1];
			platform_name[0]=0;
			clGetPlatformInfo(platformIdArray[platformIdIndex],
					CL_PLATFORM_NAME,name_size+1,platform_name,NULL);

			clGetPlatformInfo(platformIdArray[platformIdIndex],
					CL_PLATFORM_VENDOR,0,NULL,&name_size);

			char* vendor_name=new char[name_size+1];
			vendor_name[0]=0;
			clGetPlatformInfo(platformIdArray[platformIdIndex],
					CL_PLATFORM_VENDOR,name_size+1,vendor_name,NULL);

			feLog("platform %d/%d id %p vendor \"%s\" name \"%s\"\n",
					platformIdIndex,platformIdCount,
					platformIdArray[platformIdIndex],
					vendor_name,platform_name);

			//* TODO some picking logic

#if FE_GL_OPENCL
			if(platformIdIndex==platformIdCount-1)
#else
			if(platformIdIndex==0)
#endif
			{
				clPlatformId=platformIdArray[platformIdIndex];
				useGLX=(platform_name==String("NVIDIA CUDA"));
			}

			delete[] vendor_name;
			delete[] platform_name;
		}

		cl_device_type clDeviceType=
#if FE_GL_OPENCL
				useGLX? CL_DEVICES_FOR_GL_CONTEXT_KHR:
#endif
				CL_DEVICE_TYPE_ALL;

		cl_uint deviceIdCount=0;
		clGetDeviceIDs(clPlatformId,clDeviceType,
				0,NULL,&deviceIdCount);
		feLog("deviceIdCount %d\n",deviceIdCount);

		if(deviceIdCount>0)
		{
			std::vector<cl_device_id> deviceIdArray(deviceIdCount);
			cl_int clError=clGetDeviceIDs(clPlatformId,clDeviceType,
					deviceIdCount,deviceIdArray.data(),NULL);
			feLog("clGetDeviceIDs clError %d\n",clError);

			m_clDeviceId=deviceIdArray[0];

			const cl_context_properties contextProperties[]=
			{
				CL_CONTEXT_PLATFORM,
				reinterpret_cast<cl_context_properties>(clPlatformId),
				0,0
			};
			const cl_context_properties contextPropertiesGLX[]=
			{
#if FE_GL_OPENCL
				CL_GL_CONTEXT_KHR,(cl_context_properties)m_glxContext,
				CL_GLX_DISPLAY_KHR,(cl_context_properties)pDisplay,
#endif
				CL_CONTEXT_PLATFORM,
				reinterpret_cast<cl_context_properties>(clPlatformId),
				0,0
			};

			m_clContext=clCreateContext(
					useGLX? contextPropertiesGLX: contextProperties,
					deviceIdCount,deviceIdArray.data(),NULL,NULL,&clError);
			feLog("clCreateContext clError %d\n",clError);

			m_clCommandQueue=clCreateCommandQueue(m_clContext,
					m_clDeviceId,0,&clError);
			feLog("clCreateCommandQueue %p clError %d\n",
					m_clCommandQueue,clError);

			//* TEMP consider one cl_command_queue wrapped in a Safe
			m_clCommandQueue2=clCreateCommandQueue(m_clContext,
					m_clDeviceId,0,&clError);
			feLog("clCreateCommandQueue %p clError %d\n",
					m_clCommandQueue2,clError);

			registry()->master()->catalog()->catalog<void*>("clDeviceId")=
					m_clDeviceId;
			registry()->master()->catalog()->catalog<void*>("clContext")=
					m_clContext;
			registry()->master()->catalog()->catalog<void*>("clCommandQueue")=
					m_clCommandQueue;
			registry()->master()->catalog()->catalog<void*>("clCommandQueue2")=
					m_clCommandQueue2;
		}
	}
#endif

	m_spEventContextI->threadUnlock();

	setCreated(TRUE);
	return e_ok;
}

Result NativeWindow::nativeShutdown(void)
{
	if(created())
	{
		m_spEventContextI->threadLock();
		glXDestroyContext(windowHandle()->m_pDisplay,m_glxContext);
		XDestroyWindow(windowHandle()->m_pDisplay,windowHandle()->m_window);
		m_spEventContextI->threadUnlock();

		XFree(windowHandle()->m_pVisualInfo);
	}
	setCreated(FALSE);
	return e_ok;
}

void NativeWindow::setPointerMotion(BWORD active)
{
	long mask=FE_X_EVENTMASK | (active? PointerMotionMask: ButtonMotionMask);

	//* select which events we want to receive
	m_spEventContextI->threadLock();
	XSelectInput(windowHandle()->m_pDisplay,windowHandle()->m_window,mask);
	m_spEventContextI->threadUnlock();
}

void NativeWindow::setTitle(const String &title)
{
	//* set window title
	XTextProperty windowName;
	char *titles[1];
	titles[0]=(char *)title.c_str();
	if(!XStringListToTextProperty(titles,1,&windowName))
		feLog("Could not allocate space for window title");

	m_spEventContextI->threadLock();
	XSetWMName(windowHandle()->m_pDisplay,windowHandle()->m_window,&windowName);
	m_spEventContextI->threadUnlock();

	XFree((void *)windowName.value);
}

Result NativeWindow::open(const String title)
{
	if(mapped())
	{
		return e_alreadyAvailable;
	}

	startup();

	setTitle(title);

	//* map to screen and raise
	m_spEventContextI->threadLock();
	XMapRaised(windowHandle()->m_pDisplay,windowHandle()->m_window);
	XMoveWindow(windowHandle()->m_pDisplay,windowHandle()->m_window,
			m_box[0],m_box[1]);
	m_spEventContextI->threadUnlock();

	setMapped(TRUE);

	makeCurrent();
	return e_ok;
}

Result NativeWindow::close(void)
{
	m_spEventContextI->threadLock();
	XUnmapWindow(windowHandle()->m_pDisplay,windowHandle()->m_window);
	m_spEventContextI->threadUnlock();

	setMapped(FALSE);
	return e_ok;
}

void NativeWindow::makeCurrent(void)
{
	m_spEventContextI->threadLock();
	Bool success=glXMakeCurrent(windowHandle()->m_pDisplay,
			windowHandle()->m_window,m_glxContext);
	m_spEventContextI->threadUnlock();

	if(!success)
	{
		feLog("NativeWindow::makeCurrent glXMakeCurrent failed\n");
		feLog(" glGetError: \"%s\"\n",gluErrorString(glGetError()));
	}
}

void NativeWindow::releaseCurrent(void)
{
	m_spEventContextI->threadLock();
	Bool success=glXMakeCurrent(windowHandle()->m_pDisplay,None,NULL);
	m_spEventContextI->threadUnlock();

	if(!success)
	{
		feLog("NativeWindow::releaseCurrent glXMakeCurrent failed\n");
		feLog(" glGetError: \"%s\"\n",gluErrorString(glGetError()));
	}
}

void NativeWindow::swapBuffers(void)
{
	if(m_glxContext)
	{
		m_spEventContextI->threadLock();
		glXSwapBuffers(windowHandle()->m_pDisplay,windowHandle()->m_window);
		m_spEventContextI->threadUnlock();
	}
}

#endif

} /* namespace */
