/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "network.pmh"

namespace fe
{

Socket::Socket(void)
{
	m_open = false;
	m_pPoison = NULL;
	m_id = 0;
	m_nextID = 1;
}

Socket::Socket(Poison *pPoison)
{
	m_open = false;
	m_pPoison = pPoison;
	m_id = 0;
	m_nextID = 1;
}

Socket::~Socket(void)
{
	close();
}

bool Socket::isOpen(void)
{
	return m_open;
}

U32 Socket::id(void)
{
	return m_id;
}

String SockAddr::hostname(void)
{
	struct hostent *hostent_ptr;
	String name;

	hostent_ptr = gethostbyaddr((const char *)&m_sockaddr.sin_addr.s_addr,
		sizeof(struct in_addr), AF_INET);

	if (NULL == hostent_ptr)
	{
		name = "_unknown_";
	}
	else
	{
		name = (char *) hostent_ptr->h_name;
	}
	return name;
}


SockHandle Socket::socket(int domain, int type, int protocol)
{
	if((m_sock_handle = ::socket(domain, type, protocol)) < 0)
	{
		feX(e_system,
			"fe::Socket::socket",
			errorString(FE_ERRNO).c_str());
	}
	return m_sock_handle;
}

void Socket::openStream(void)
{
	socket(AF_INET, SOCK_STREAM, 0);
}

void Socket::bind(SockAddr &sockaddr)
{
	if(::bind(m_sock_handle, sockaddr.ptr(), sockaddr.size()) < 0)
	{
		feLogGroup("network","Socket::bind failed\n");
		feX(e_system,
			"fe::Socket::bind",
			errorString(FE_ERRNO).c_str());
	}
}

void Socket::listen(int backlog)
{
	if(::listen(m_sock_handle, backlog) < 0)
	{
		feX(e_system,
			"fe::Socket::listen",
			errorString(FE_ERRNO).c_str());
	}
}

sp<Socket> Socket::accept(SockAddr &sockaddr)
{
	sp<Socket> new_sock(new Socket);
	while(true)
	{
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		int size;
		size = sockaddr.size();
#else
		socklen_t size;
		size = (socklen_t)(sockaddr.size());
#endif
		wait();
		if((new_sock->m_sock_handle =
			::accept(m_sock_handle, sockaddr.ptr(), &size) ) < 0)
		{
			if(FE_ERRNO == FE_EINTR)
			{
				continue;
			}
			else
			{
				feX(e_system,
					"fe::Socket::accept",
					errorString(FE_ERRNO).c_str());
			}
		}
		else
		{
			new_sock->m_open = true;
			U32 net_id;
			{
				SAFEGUARD;
				net_id = htonl(m_nextID);
				new_sock->m_id = m_nextID;
				m_nextID++;
			}
			int n = new_sock->write((void *)(&net_id), sizeof(U32));
			if(n != sizeof(U32))
			{
				feX(e_system,
					"fe::Socket::accept",
					errorString(FE_ERRNO).c_str());
			}
			return new_sock;
		}
	}
}

void Socket::connect(SockAddr &sockaddr)
{
	if(::connect(m_sock_handle, sockaddr.ptr(), sockaddr.size()) < 0)
	{
		feX(e_system,
			"fe::Socket::connect",
			errorString(FE_ERRNO).c_str());
	}
	m_open = true;
	U32 net_id;
	int n = read((void *)(&net_id), sizeof(U32));
	if(n != sizeof(U32))
	{
		feX(e_system,
			"fe::Socket::connect",
			errorString(FE_ERRNO).c_str());
	}
	m_id = ntohl(net_id);
#if FE_CODEGEN<=FE_DEBUG
	feLogGroup("network","connect id %u\n", m_id);
#endif
}

void Socket::close(void)
{
	if(m_open)
	{
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
		if(closesocket(m_sock_handle) != 0)
#else
		if(::close(m_sock_handle) < 0)
#endif
		{
			feX(e_system,
				"fe::Socket::close",
				errorString(FE_ERRNO).c_str());
		}
		m_open = false;
	}
}

void Socket::setPoison(Poison *pPoison)
{
	m_pPoison = pPoison;
}

void Socket::wait(void)
{
	if(m_pPoison == NULL)
	{
#if FE_CODEGEN<=FE_DEBUG
		feLogGroup("network","NULL poison\n");
#endif

		return;
	}
	fd_set rfds;
	struct timeval tv;
	int ready = 0;
	while(!ready)
	{
		if(m_pPoison->active())
		{
			close();

#if FE_CODEGEN<=FE_DEBUG
			feLogGroup("network","Socket throwing poison\n");
#endif

			throw Poisoned();
		}
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		FD_SET(m_sock_handle, &rfds);
		ready = select(m_sock_handle+1, &rfds, NULL, NULL, &tv);
	}
}

IWORD Socket::read(const void *pBuffer, IWORD count)
{
	if(!m_open)
	{
		feX(e_readFailed,
			"fe::Socket::read",
			"attempt to read from closed socket");
	}
	IWORD nleft;
	IWORD nread;
	char *ptr = (char *)pBuffer;
	nleft = count;
	while(nleft > 0)
	{
		wait();
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64 || FE_OS==FE_OSX
		if(( nread = ::recv(m_sock_handle, ptr, nleft,0) ) < 0 )
#else
		if(( nread = ::recv(m_sock_handle, ptr, nleft,MSG_NOSIGNAL) ) < 0 )
#endif
		{
			if(FE_ERRNO == FE_EINTR)
			{
				nread = 0;
			}
			else
			{
				feLogGroup("network","%u %s\n", FE_ERRNO,
						errorString(FE_ERRNO).c_str());
				return -1;
			}
		}
		else if(nread == 0)
		{
			break;
		}
		nleft -= nread;
		ptr += nread;
	}
	return count - nleft;
}

IWORD Socket::write(const void *pBuffer, IWORD count)
{
	if(!m_open)
	{
		feX(e_writeFailed,
			"fe::Socket::write",
			"attempt to write to closed socket");
	}
	IWORD nleft;
	IWORD nwritten;
	const char *ptr;
	ptr = (const char *)pBuffer;
	nleft = count;
	while(nleft > 0)
	{
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64 || FE_OS==FE_OSX
		if((nwritten = ::send(m_sock_handle, ptr, nleft,0)) <= 0)
#else
		if((nwritten = ::send(m_sock_handle, ptr, nleft,MSG_NOSIGNAL)) <= 0)
#endif
		{
			switch(FE_ERRNO)
			{
				case FE_EINTR:
					nwritten = 0;
					break;
				default:
					feLogGroup("network","%u %s\n", FE_ERRNO,
							errorString(FE_ERRNO).c_str());
					return -1;
			}
		}
		nleft -= nwritten;
		ptr += nwritten;
	}
	return count;
}

void Socket::startup(void)
{
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	WORD wVersionRequested = MAKEWORD(2,2);
	WSADATA wsaData;
	int err = WSAStartup(wVersionRequested, &wsaData);

	if(0 != err)
	{
		feX(e_system,
				"fe::Socket::startup",
				errorString(FE_ERRNO).c_str());
	}

	if(2 != LOBYTE(wsaData.wVersion) ||
		2 != HIBYTE(wsaData.wVersion))
	{
		WSACleanup();
		feX(e_system,
				"fe::Socket::startup",
				errorString(FE_ERRNO).c_str());
	}
#endif
}

void Socket::shutdown(void)
{
#if FE_OS==FE_WIN32 || FE_OS==FE_WIN64
	int err = WSACleanup();
	if(err)
	{
		feX(e_system,
				"fe::Socket::shutdown",
				errorString(FE_ERRNO).c_str());
	}
#endif
}


} /* namespace */

