/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

#define FE_SAC_DEBUG	FALSE

namespace fe
{

void SurfaceAttrCreateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<String>("name")="attribute";
	catalog<String>("name","label")="Name";

	catalog<String>("class")="point";
	catalog<String>("class","choice:0")="point";
	catalog<String>("class","label:0")="Point";
	catalog<String>("class","choice:1")="primitive";
	catalog<String>("class","label:1")="Primitive";
	catalog<String>("class","label")="Class";

	catalog<String>("type")="real";
	catalog<String>("type","choice:0")="real";
	catalog<String>("type","label:0")="Real";
	catalog<String>("type","choice:1")="integer";
	catalog<String>("type","label:1")="Integer";
	catalog<String>("type","choice:2")="vector";
	catalog<String>("type","label:2")="Vector";
	catalog<String>("type","choice:3")="string";
	catalog<String>("type","label:3")="String";
	catalog<String>("type","label")="Type";

	catalog<String>("default");
	catalog<String>("default","label")="Default";

	catalog<bool>("writeValues")=false;
	catalog<String>("writeValues","label")="Write Values";
	catalog<bool>("writeValues","joined")=true;

	catalog<String>("value");
	catalog<String>("value","label")="Value";
	catalog<String>("value","enabler")="writeValues";

	catalog<bool>("rampValue")=false;
	catalog<String>("rampValue","label")="Ramp";
	catalog<bool>("rampValue","joined")=true;

	catalog<String>("rampTo");
	catalog<String>("rampTo","label")="To";
	catalog<String>("rampTo","enabler")="rampValue";
	catalog<bool>("rampTo","joined")=true;

	catalog<Real>("rampGamma")=1.0;
	catalog<String>("rampGamma","label")="Gamma";
	catalog<String>("rampGamma","enabler")="rampValue";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","clobber")=true;
}

void SurfaceAttrCreateOp::handle(Record& a_rSignal)
{
#if FE_SAC_DEBUG
	feLog("SurfaceAttrCreateOp::handle\n");
#endif

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	const String attrName=catalog<String>("name");
	if(attrName.empty())
	{
		feLog("SurfaceAttrCreateOp::handle no name\n");
		return;
	}
	const String attrClass=catalog<String>("class");
	const String attrType=catalog<String>("type");
	const String& attrDefault=catalog<String>("default");

	const BWORD attrWriteValues=catalog<bool>("writeValues");
	const String& attrValue=catalog<String>("value");

	const BWORD rampValue=catalog<bool>("rampValue");
	const String& rampTo=catalog<String>("rampTo");
	const Real rampGamma=catalog<Real>("rampGamma");

	sp<SurfaceAccessorI> spOutputElement;
	access(spOutputElement,spOutputAccessible,
			attrClass=="point"? e_point: e_primitive,
			attrName,e_quiet,e_refuseMissing);
	if(spOutputElement.isValid())
	{
		if(!attrWriteValues)
		{
			//* don't clobber existing data
			return;
		}
	}
	else
	{
		//* create new attribute
		if(!access(spOutputElement,spOutputAccessible,
				attrClass=="point"? e_point: e_primitive,attrName)) return;
	}

	const String& attrSet=(attrWriteValues)? attrValue: attrDefault;

	I32 elementCount=spOutputElement->count();

#if FE_SAC_DEBUG
	feLog("SurfaceAttrCreateOp::handle elementCount=%d\n",elementCount);
#endif

	if(attrType=="string")
	{
		for(I32 index=0;index<elementCount;index++)
		{
#if FE_SAC_DEBUG
			feLog("SurfaceAttrCreateOp::handle set %d \"%s\"\n",
					index,attrSet.c_str());
#endif
			spOutputElement->set(index,0,attrSet);
		}
	}
	else if(attrType=="integer")
	{
		const I32 value=atoi(attrSet.c_str());
		for(I32 index=0;index<elementCount;index++)
		{
#if FE_SAC_DEBUG
			feLog("SurfaceAttrCreateOp::handle set %d %d\n",
					index,value);
#endif
			spOutputElement->set(index,0,value);
		}
	}
	else if(attrType=="real")
	{
		const Real value=atof(attrSet.c_str());

		if(rampValue)
		{
			const Real rampRange=atof(rampTo.c_str())-value;

			for(I32 index=0;index<elementCount;index++)
			{
				const Real ramped=value+
						rampRange*pow((index/Real(elementCount-1)),rampGamma);

#if FE_SAC_DEBUG
				feLog("SurfaceAttrCreateOp::handle set %d %.6G\n",
						index,ramped);
#endif

				spOutputElement->set(index,0,ramped);
			}
		}
		else
		{
			for(I32 index=0;index<elementCount;index++)
			{
#if FE_SAC_DEBUG
				feLog("SurfaceAttrCreateOp::handle set %d %.6G\n",
						index,value);
#endif
				spOutputElement->set(index,0,value);
			}
		}
	}
	else if(attrType=="vector")
	{
		Real value0;
		Real value1;
		Real value2;

		sscanf(attrSet.c_str(),"%f %f %f",&value0,&value1,&value2);

		const SpatialVector defaultVector(value0,value1,value2);

		for(I32 index=0;index<elementCount;index++)
		{
#if FE_SAC_DEBUG
			feLog("SurfaceAttrCreateOp::handle set %d %s\n",
					index,c_print(defaultVector));
#endif
			spOutputElement->set(index,0,defaultVector);
		}
	}

#if FE_SAC_DEBUG
	feLog("SurfaceAttrCreateOp::handle done\n");
#endif
}

} // namespace
