/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "hydra.pmh"

#define FE_DHY_DEBUG			FALSE

namespace fe
{

sp<DrawBufferI> DrawHydra::createBuffer(void)
{
	sp<DrawHydra::Buffer> spDrawBuffer(new DrawHydra::Buffer);

	if(spDrawBuffer.isValid())
	{
		static I32 serial=1;	//* HACK

		String bufferName;
		bufferName.sPrintf("temp%06d",serial++);
		spDrawBuffer->setName(bufferName);
	}

	return spDrawBuffer;
}

DrawHydra::Buffer::Buffer(void):
	m_pHydraDriver(NULL),
	m_current(FALSE),
	m_recent(0),
	m_multicolor(FALSE),
	m_firstColor(0,0,0,0)
{
}

DrawHydra::Buffer::~Buffer(void)
{
	releaseAll();

	if(m_pHydraDriver)
	{
		if(!m_nodeName.empty())
		{
//			feLog("DROP \"%s\"\n",m_nodeName.c_str());

			const SdfPath id(("/"+m_nodeName).c_str());
			m_pHydraDriver->GetDelegate().Remove(id);
		}

		m_pHydraDriver=NULL;
	}
}

void DrawHydra::Buffer::releaseAll(void)
{
	m_current=FALSE;

	DrawOpenGL::Buffer::releaseAll();
}

DrawHydra::DrawHydra(void):
	m_pHydraDriver(NULL)
{
	m_drawn=FALSE;
	m_refineLevel=0;
	m_alphaMin=0.1;				//* TODO param
	m_shadowResolution=32768;	//* TODO param
	m_shadowBias= -0.001;		//* TODO param
}

DrawHydra::~DrawHydra(void)
{
}

void DrawHydra::initialize(void)
{
#if FE_DHY_DEBUG
	feLog("DrawHydra::initialize\n");
#endif

	const char* shading=(const char *)glGetString(GL_SHADING_LANGUAGE_VERSION);
	const Real version=atof(shading);
	if(version<Real(4))
	{
		feLog("DrawHydra::initialize GLSL 4 required, found \"%s\"\n",shading);

		feX(e_cannotCreate,"DrawHydra::initialize","hardware not capable");
	}
}

void DrawHydra::flush(void)
{
#if FE_DHY_DEBUG
//	feLog("DrawHydra::flush %p\n",m_pHydraDriver);
#endif

	DrawOpenGL::flush();

	m_drawn=FALSE;

//	updateNodes();
}

template <typename T>
static VtArray<T> _BuildArray(T values[], int numValues)
{
	VtArray<T> result(numValues);
	std::copy(values, values+numValues, result.begin());
	return result;
}

void DrawHydra::ensureDriver(void)
{
	if(!m_pHydraDriver)
	{
		m_pHydraDriver = new HydraDriver(HdTokens->hull);

		HdRenderPassStateSharedPtr const& rRenderPassState=
				m_pHydraDriver->GetRenderPassState();

		//* NOTE otherwise, alpha below 0.5 just disappears
		rRenderPassState->SetAlphaToCoverageUseDefault(false);
//		rRenderPassState->SetAlphaToCoverageEnabled(false);
		rRenderPassState->SetAlphaThreshold(0.01);

//		rRenderPassState->SetCullStyle(HdCullStyleFront);
		rRenderPassState->SetCullStyle(HdCullStyleNothing);

		HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();

		m_refineLevel=drawMode()->refinement();
		rDelegate.SetRefineLevel(m_refineLevel);

		rDelegate.AddSimpleLightTask(SdfPath("/SimpleLightTask"));
		rDelegate.AddShadowTask(SdfPath("/ShadowTask"));
		rDelegate.AddRenderSetupTask(SdfPath("/RenderSetupTask"));
		rDelegate.AddRenderTask(SdfPath("/RenderTask"));
	}
}

void DrawHydra::updateLighting(void)
{
	sp<CameraDisplay> spCameraDisplay=view()->cameraDisplay();
	const Box2i viewport=spCameraDisplay->viewport();

	Matrix<4,4,Real> proj4x4;
	if(spCameraDisplay->projection()==CameraI::e_perspective)
	{
		Real fovy;
		Real nearplane;
		Real farplane;
		spCameraDisplay->getPerspective(fovy,nearplane,farplane);

		const GLfloat aspect=(GLfloat)width(viewport)/
				(GLfloat)height(viewport);
		proj4x4=perspective(fovy,Real(aspect),nearplane,farplane);
	}
	else
	{
		Real zoom;
		Vector2 center;
		spCameraDisplay->getOrtho(zoom,center);

		const Vector2 size(width(viewport),height(viewport));
		proj4x4=ViewCommon::computeOrtho(zoom,center,size);
	}

	const GfMatrix4d projMatrix(
			proj4x4[0],proj4x4[1],proj4x4[2],proj4x4[3],
			proj4x4[4],proj4x4[5],proj4x4[6],proj4x4[7],
			proj4x4[8],proj4x4[9],proj4x4[10],proj4x4[11],
			proj4x4[12],proj4x4[13],proj4x4[14],proj4x4[15]);

	Real view16[16];
	spCameraDisplay->cameraMatrix().copy16(view16);

	const GfMatrix4d viewMatrix(
			view16[0],view16[1],view16[2],view16[3],
			view16[4],view16[5],view16[6],view16[7],
			view16[8],view16[9],view16[10],view16[11],
			view16[12],view16[13],view16[14],view16[15]);

	m_pHydraDriver->SetCamera(viewMatrix, projMatrix,
			GfVec4d(0, 0, width(viewport), height(viewport)));

	SpatialTransform cameraTransform;
	invert(cameraTransform,spCameraDisplay->cameraMatrix());

	const SpatialVector lightDirFromCamera(-1.0,-1.0,-3.0);
	SpatialVector lightDir;
	rotateVector(cameraTransform,lightDirFromCamera,lightDir);
	normalizeSafe(lightDir);

	m_brightness=fe::minimum(Real(1),fe::maximum(Real(0),m_brightness));
	m_contrast=fe::minimum(Real(2),fe::maximum(Real(0),m_contrast));

//	m_lightingShader->SetLightingStateFromOpenGL();

	GlfSimpleLightVector lights(1);
	GlfSimpleLight& rLight=lights[0];
	rLight.SetPosition(GfVec4f(-lightDir[0],-lightDir[1],-lightDir[2],0));
	rLight.SetAmbient(GfVec4f(m_brightness));
	rLight.SetDiffuse(GfVec4f(m_contrast,m_contrast,m_contrast,1));
	rLight.SetSpecular(GfVec4f(0.5));
	rLight.SetShadowResolution(m_shadowResolution);
	rLight.SetShadowBias(m_shadowBias);
//	rLight.SetShadowBlur(GfVec4f(0.1));

	rLight.SetHasShadow(drawMode()->shadows());

	const SdfPath lightPath("/Light0");

	HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();
	if(!rDelegate.HasLight(lightPath))
	{
		rDelegate.AddLight(lightPath,rLight);
	}
	rDelegate.SetLight(lightPath,HdStLightTokens->params,
			VtValue(rLight));

	const SdfPath renderSetupPath("/RenderSetupTask");
	rDelegate.AdjustRenderSetupTask(renderSetupPath,drawMode()->lit(),
			GfVec4f(0, 0, width(viewport), height(viewport)));
}

void DrawHydra::drawPoints(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	BWORD multicolor,const Color *color,
	sp<DrawBufferI> spDrawBuffer)
{
//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawPoints vertices %d buffer %p\n",
//				vertices,spDrawBuffer.raw());
//	}

	if(spDrawBuffer.isNull())
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawPoints bypass %d\n",vertices);
//		}

		DrawOpenGL::drawPoints(vertex,normal,vertices,
				multicolor,color,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawPoints incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	ensureDriver();

	if(!spBuffer->m_current)
	{
		const I32 refinement=drawMode()->refinement();
		const BWORD refined=(refinement>0);

		HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();

		spBuffer->m_pHydraDriver=m_pHydraDriver;
		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

		const I32 pointCount=vertices;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPoints hull points %d refine %d\n",
				pointCount,refinement);
#endif

		GfVec3f* points=new GfVec3f[pointCount];

		//* NOTE Hydra does not currently seem to care about point normals
		//* (uses normals always facing camera)
		GfVec3f* normals=NULL;
		if(FALSE&& normal)
		{
			normals=new GfVec3f[pointCount];
		}

		HydraDelegate::Interpolation colorInterp=HydraDelegate::CONSTANT;
		VtValue vtColor;
		VtVec4fArray vtColorArray;
		if(multicolor)
		{
			colorInterp=HydraDelegate::VERTEX;
			vtColorArray.resize(pointCount);
		}
		else
		{
			const Color& rColor=color[0];

			vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
					fe::maximum(Real(rColor[3]),m_alphaMin));
		}

		SpatialVector min(FLT_MAX,FLT_MAX,FLT_MAX);
		SpatialVector max(FLT_MIN,FLT_MIN,FLT_MIN);

		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			const SpatialVector& point=vertex[pointIndex];

			points[pointIndex].Set(point.temp());

			for(I32 m=0;m<3;m++)
			{
				if(min[m]>point[m])
				{
					min[m]=point[m];
				}
				if(max[m]<point[m])
				{
					max[m]=point[m];
				}
			}

			if(normals)
			{
				normals[pointIndex].Set(normal[pointIndex].temp());
			}

			if(multicolor)
			{
				const Color& rColor=color[pointIndex];
				Real alpha=rColor[3];

#if TRUE
				//* HACK boost alpha
				alpha=pow(alpha,0.3);
				if(alpha>0.01)
				{
					alpha=fe::maximum(m_alphaMin,fe::minimum(alpha,Real(1)));
				}
#endif

				vtColorArray[pointIndex]=
						GfVec4f(rColor[0],rColor[1],rColor[2],alpha);
			}
		}

		const SpatialVector size=max-min;
		const Real largest=fe::maximum(size[0],fe::maximum(size[1],size[2]));
		const Real defaultWidth=0.01*largest;	//* NOTE tweak

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		if(multicolor)
		{
			vtColor = VtValue(vtColorArray);
		}

		if(spBuffer->m_nodeName.empty())
		{
			spBuffer->m_nodeName=
					spBuffer->name().substitute(":","_").substitute(".","_");
		}

		const SdfPath id(("/"+spBuffer->m_nodeName).c_str());

#if TRUE
		const VtValue vtWidth((float)defaultWidth);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::CONSTANT;
#else
		VtFloatArray vtWidthArray;
		vtWidthArray.resize(pointCount);
		for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
		{
			vtWidthArray[pointIndex]=(pointIndex%2)? 0.01f: 0.04f;
		}
		const VtValue vtWidth=VtValue(vtWidthArray);
		const HydraDelegate::Interpolation widthInterp=HydraDelegate::VERTEX;
#endif

		rDelegate.Remove(id);
		m_bufferMap.erase(spBuffer->m_nodeName);

		SdfPath instancerId;
		rDelegate.AddPoints(id,
				_BuildArray(points,pointCount),
				normalArray,
				vtColor,colorInterp,
				vtWidth,widthInterp,
				instancerId);

		//* NOTE point modes are all equivalent
		rDelegate.SetReprName(id, refined? HdTokens->refined: HdTokens->hull);

		rDelegate.SetRefineLevel(id,drawMode()->refinement());

		spBuffer->m_current=TRUE;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPoints new\n");
#endif

		delete[] points;
		points=NULL;
	}

	//* HACK high resolution shadows don't seem to work well for point clouds
	m_shadowBias= -0.1;
	m_shadowResolution=256;

	spBuffer->m_recent=1;

	m_bufferMap[spBuffer->m_nodeName]=spBuffer;

	updateNodes();
}

void DrawHydra::drawTriangles(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	if(drawStyle()==DrawMode::e_pointCloud || spDrawBuffer.isNull() ||
			!hullPointMap || !hullFacePoint || isUvSpace())
	{
		DrawOpenGL::drawTriangles(vertex,normal,texture,vertices,
				strip,multicolor,color,
				hullPointMap,hullFacePoint,spDrawBuffer);
		return;
	}

	drawPolys(3,vertex,normal,texture,vertices,strip,
			multicolor,color,hullPointMap,hullFacePoint,spDrawBuffer);
}

void DrawHydra::drawLinesInternal(const SpatialVector* vertex,
	const SpatialVector* normal,const Vector2f* texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color,
	BWORD multiradius,const Real* radius,
	const Vector3i* element,U32 elementCount,
	sp<DrawBufferI> spDrawBuffer)
{
//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawLinesInternal"
//				" vertices %d buffer %p elementCount %d\n  %s\n",
//				vertices,spDrawBuffer.raw(),elementCount,
//				c_print(drawMode()));
//	}

	if(strip==e_strip && !element)
	{
		elementCount=1;
	}

	const BWORD uvSpace=isUvSpace();
	const BWORD capable=(elementCount>0);

	//* HACK
//	if(capable && spDrawBuffer.isNull())
//	{
//		spDrawBuffer=createBuffer();
//	}

	if(!capable || spDrawBuffer.isNull())
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawLinesInternal bypass %d\n",vertices);
//		}

		DrawOpenGL::drawLinesInternal(vertex,normal,texture,vertices,strip,
				multicolor,color,multiradius,radius,
				element,elementCount,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawLinesInternal incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	ensureDriver();

	if(!spBuffer->m_current)
	{
		if(uvSpace && !texture)
		{
			feLog("DrawHydra::drawLinesInternal"
					" vertices %d in UV space without UV coordinates\n",
					vertices);
			return;
		}

		HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();

		spBuffer->m_pHydraDriver=m_pHydraDriver;
		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

//		const GfVec3f center=rDelegate.PopulateBasicTestSet();

		const I32 pointCount=vertices+2*elementCount;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawLinesInternal"
				" hull points %d faces %d refine %d\n",
				pointCount,elementCount,drawMode()->refinement());
#endif

		GfVec3f* points=new GfVec3f[pointCount];
		int* numVerts=new int[elementCount];
		I32 pointIndex=0;

		GfVec3f* normals=NULL;
		if(FALSE&& normal)
		{
			normals=new GfVec3f[pointCount];
		}

		HydraDelegate::Interpolation colorInterp=HydraDelegate::CONSTANT;
		VtValue vtColor;
		VtVec4fArray vtColorArray;
		if(multicolor)
		{
			colorInterp=HydraDelegate::VERTEX;
			vtColorArray.resize(pointCount);
		}
		else
		{
			const Color& rColor=color[0];

			vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
					fe::maximum(Real(rColor[3]),m_alphaMin));
		}

		HydraDelegate::Interpolation widthInterp=HydraDelegate::CONSTANT;
		VtValue vtWidth;
		VtFloatArray vtWidthArray;
		if(multiradius)
		{
			widthInterp=HydraDelegate::VERTEX;
			vtWidthArray.resize(pointCount);
		}
		else if(radius)
		{
			vtWidth=float(radius[0]);
		}

		SpatialVector min(FLT_MAX,FLT_MAX,FLT_MAX);
		SpatialVector max(FLT_MIN,FLT_MIN,FLT_MIN);

		for(U32 elementIndex=0;elementIndex<elementCount;elementIndex++)
		{
			I32 startIndex=0;
			I32 subCount=vertices;

			if(element)
			{
				const Vector3i& rElement=element[elementIndex];
				startIndex=rElement[0];
				subCount=rElement[1];
			}

			const I32 cvCount=subCount+2;

			for(I32 subIndex=0;subIndex<subCount;subIndex++)
			{
				SpatialVector point;
				if(uvSpace)
				{
					const Vector2f& rUV=texture[startIndex+subIndex];
					set(point,rUV[0],rUV[1],Real(0));
				}
				else
				{
					point=vertex[startIndex+subIndex];
				}

				points[pointIndex+1+subIndex].Set(point.temp());

				for(I32 m=0;m<3;m++)
				{
					if(min[m]>point[m])
					{
						min[m]=point[m];
					}
					if(max[m]<point[m])
					{
						max[m]=point[m];
					}
				}
			}
			points[pointIndex]=points[pointIndex+1];
			points[pointIndex+cvCount-1]=points[pointIndex+cvCount-2];

			if(normals)
			{
				for(I32 subIndex=0;subIndex<subCount;subIndex++)
				{
					//* NOTE normals appear to be in eye space
//					normals[pointIndex+1+subIndex].Set(
//							normal[startIndex+subIndex].temp());
					normals[pointIndex+1+subIndex]=
							GfVec3f(0.0f,0.0f,1.0f);
				}
				normals[pointIndex]=normals[pointIndex+1];
				normals[pointIndex+cvCount-1]=
						normals[pointIndex+cvCount-2];
			}

			if(multicolor)
			{
				for(I32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const Color& rColor=color[startIndex+subIndex];

					vtColorArray[pointIndex+1+subIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(Real(rColor[3]),m_alphaMin));
				}
				vtColorArray[pointIndex]=vtColorArray[pointIndex+1];
				vtColorArray[pointIndex+cvCount-1]=
						vtColorArray[pointIndex+cvCount-2];
			}

			if(multiradius)
			{
				for(I32 subIndex=0;subIndex<subCount;subIndex++)
				{
					const Real& rRadius=radius[startIndex+subIndex];

					vtWidthArray[pointIndex+1+subIndex]=float(2.0*rRadius);
				}
				vtWidthArray[pointIndex]=vtWidthArray[pointIndex+1];
				vtWidthArray[pointIndex+cvCount-1]=
						vtWidthArray[pointIndex+cvCount-2];
			}

			numVerts[elementIndex]=cvCount;
			pointIndex+=cvCount;
		}

		const SpatialVector size=max-min;
		const Real largest=fe::maximum(size[0],fe::maximum(size[1],size[2]));
		const Real defaultWidth=2e-4*largest;	//* NOTE tweak

		FEASSERT(pointIndex==pointCount);

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		if(multicolor)
		{
			vtColor = VtValue(vtColorArray);
		}

		if(multiradius)
		{
			vtWidth = VtValue(vtWidthArray);
		}
		else if(!radius)
		{
			vtWidth = float(defaultWidth);
		}

		if(spBuffer->m_nodeName.empty())
		{
			spBuffer->m_nodeName=
					spBuffer->name().substitute(":","_").substitute(".","_");
		}

		const SdfPath id(("/"+spBuffer->m_nodeName).c_str());
		const TfToken basis=HdTokens->catmullRom;

		rDelegate.Remove(id);
		m_bufferMap.erase(spBuffer->m_nodeName);

		SdfPath instancerId;
		rDelegate.AddBasisCurves(id,
				_BuildArray(points,pointCount),
				_BuildArray(numVerts,elementCount),
				normalArray,
				basis,
				vtColor,colorInterp,
				vtWidth,widthInterp,
				instancerId);

		const BWORD refined=(drawMode()->refinement()>0);

		//* NOTE choices:
		//*		hull (same as smoothHull, wire, wireOnSurf, refinedWire)
		//*		refined (same as refinedWireOnSurf)
		//* see: pxr/imaging/lib/hd/renderIndex.cpp (all same for points)
		rDelegate.SetReprName(id, refined? HdTokens->refined: HdTokens->hull);

		rDelegate.SetRefineLevel(id,drawMode()->refinement());

		spBuffer->m_current=TRUE;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawLinesInternal new\n");
#endif

		delete[] numVerts;
		numVerts=NULL;

		delete[] points;
		points=NULL;
	}

	spBuffer->m_recent=1;

	m_bufferMap[spBuffer->m_nodeName]=spBuffer;

	updateNodes();
}

void DrawHydra::drawPolys(U32 grain,const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,
	U32 vertices,StripMode strip,BWORD multicolor,const Color *color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
	const BWORD hasHull=(hullPointMap && hullFacePoint);

//	if(vertices>8)
//	{
//		feLog("DrawHydra::drawPolys"
//				" vertices %d grain %d hull %d buffer %p\n  %s\n",
//				vertices,grain,hasHull,spDrawBuffer.raw(),c_print(drawMode()));
//	}

	const BWORD uvSpace=isUvSpace();
	const BWORD capable=(!uvSpace && (grain==3 || grain==4));

	//* HACK
	if(capable && spDrawBuffer.isNull())
	{
#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPolys createBuffer\n");
#endif
		spDrawBuffer=createBuffer();
	}

	if(!capable || spDrawBuffer.isNull())
	{
//		if(vertices>8)
//		{
//			feLog("DrawHydra::drawPolys bypass %d\n",vertices);
//		}

		DrawOpenGL::drawPolys(grain,vertex,normal,texture,vertices,strip,
				multicolor,color,hullPointMap,hullFacePoint,spDrawBuffer);
		return;
	}

	sp<Buffer> spBuffer=spDrawBuffer;
	if(spBuffer.isNull())
	{
		feLog("DrawHydra::drawPolys incompatible buffer\n");
		return;
	}

	if(spBuffer->m_multicolor!=multicolor || spBuffer->m_firstColor!=color[0])
	{
		spBuffer->releaseAll();
	}

	ensureDriver();

	HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();

	if(!spBuffer->m_current)
	{
		if(uvSpace && !texture)
		{
			feLog("DrawHydra::drawPolys"
					" vertices %d in UV space without UV coordinates\n",
					vertices);
			return;
		}

		//* NOTE no refinement without hull
		const I32 refinement=hasHull? drawMode()->refinement(): 0;
		const BWORD refined=(refinement>0);

		spBuffer->m_pHydraDriver=m_pHydraDriver;
		spBuffer->m_multicolor=multicolor;
		spBuffer->m_firstColor=color[0];

//		const GfVec3f center=rDelegate.PopulateBasicTestSet();

		// VERTEX, UNIFORM, CONSTANT, FACEVARYING, VARYING
		const HydraDelegate::Interpolation interpolation=multicolor?
				HydraDelegate::VERTEX:
				HydraDelegate::CONSTANT;

		I32 pointCount=0;
		I32 faceCount=0;
		GfVec3f* points=NULL;
		int* numVerts=NULL;
		int* verts=NULL;
		int vertCount=0;
		GfVec3f* normals=NULL;
		BWORD allTriangles=TRUE;
		VtValue vtColor;
		VtVec4fArray vtColorArray;

		TfToken orientation = HdTokens->leftHanded;
		I32 pointIndex0= -1;
		I32 pointIndex1= -1;
		I32 pointIndex2= -1;

		if(hasHull)
		{
			const Array<I32>& rHullPointMap= *hullPointMap;
			const Array<Vector4i>& rHullFacePoint= *hullFacePoint;

			pointCount=rHullPointMap.size();
			faceCount=rHullFacePoint.size();

#if FE_DHY_DEBUG
			feLog("DrawHydra::drawPolys hull points %d faces %d\n",
					pointCount,faceCount);
#endif

			points=new GfVec3f[pointCount];
			numVerts=new int[faceCount];
			verts=new int[faceCount*4];

			if(uvSpace)
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Vector2f& rUV=texture[rHullPointMap[pointIndex]];
					points[pointIndex].Set(rUV[0],rUV[1],Real(0));
				}
			}
			else
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					points[pointIndex].Set(
							vertex[rHullPointMap[pointIndex]].temp());
				}
			}

			//* NOTE refinement generates normals (Hydra will complain)
			if(normal)
			{
				if(faceCount)
				{
					const Vector4i& rFacePoint=rHullFacePoint[0];
					pointIndex0=rHullPointMap[rFacePoint[0]];
					pointIndex1=rHullPointMap[rFacePoint[1]];
					pointIndex2=rHullPointMap[rFacePoint[2]];
				}

				if(!refined)
				{
					normals=new GfVec3f[pointCount];

					for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
					{
						normals[pointIndex].Set(
								normal[rHullPointMap[pointIndex]].temp());
					}
				}
			}

			for(I32 faceIndex=0;faceIndex<faceCount;faceIndex++)
			{
				const Vector4i& rFacePoint=rHullFacePoint[faceIndex];
				if(rFacePoint[3]<0)
				{
					numVerts[faceIndex]=3;
					verts[vertCount++]=rFacePoint[0];
					verts[vertCount++]=rFacePoint[1];
					verts[vertCount++]=rFacePoint[2];
				}
				else
				{
					numVerts[faceIndex]=4;
					verts[vertCount++]=rFacePoint[0];
					verts[vertCount++]=rFacePoint[1];
					verts[vertCount++]=rFacePoint[2];
					verts[vertCount++]=rFacePoint[3];
					allTriangles=FALSE;
				}
			}

			const BWORD boostAlpha=(drawStyle()==DrawMode::e_wireframe);

			if(multicolor)
			{
				vtColorArray.resize(pointCount);
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Color& rColor=color[rHullPointMap[pointIndex]];
					Real alpha=rColor[3];
					if(boostAlpha)
					{
						alpha=pow(alpha,0.5);
					}

					vtColorArray[pointIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(alpha,m_alphaMin));
				}
				vtColor = VtValue(vtColorArray);
			}
			else
			{
				const Color& rColor=color[0];
				Real alpha=rColor[3];
				if(boostAlpha)
				{
					alpha=pow(alpha,0.5);
				}

				vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
						fe::maximum(alpha,m_alphaMin));
			}
		}
		else
		{
			pointCount=vertices;
			faceCount=vertices/grain;

#if FE_DHY_DEBUG
			feLog("DrawHydra::drawPolys discrete points %d faces %d\n",
					pointCount,faceCount);
#endif

			points=new GfVec3f[pointCount];
			numVerts=new int[faceCount];
			verts=new int[pointCount];

			if(uvSpace)
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Vector2f& rUV=texture[pointIndex];
					points[pointIndex].Set(rUV[0],rUV[1],Real(0));
					verts[pointIndex]=pointIndex;
				}
			}
			else
			{
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					points[pointIndex].Set(vertex[pointIndex].temp());
					verts[pointIndex]=pointIndex;
				}
			}

			if(normal)
			{
				normals=new GfVec3f[pointCount];

				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					normals[pointIndex].Set(normal[pointIndex].temp());
				}

				pointIndex0=0;
				pointIndex1=1;
				pointIndex2=2;
			}

			vertCount=pointCount;
			allTriangles=(grain==3);

			for(I32 faceIndex=0;faceIndex<faceCount;faceIndex++)
			{
				numVerts[faceIndex]=grain;
			}

			if(multicolor)
			{
				vtColorArray.resize(pointCount);
				for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
				{
					const Color& rColor=color[pointIndex];

					vtColorArray[pointIndex]=GfVec4f(
							rColor[0],rColor[1],rColor[2],
							fe::maximum(Real(rColor[3]),m_alphaMin));
				}
				vtColor = VtValue(vtColorArray);
			}
			else
			{
				const Color& rColor=color[0];

				vtColor=GfVec4f(rColor[0],rColor[1],rColor[2],
						fe::maximum(Real(rColor[3]),m_alphaMin));
			}
		}

		if(pointIndex0>=0)
		{
			const SpatialVector& p0=vertex[pointIndex0];
			const SpatialVector& p1=vertex[pointIndex1];
			const SpatialVector& p2=vertex[pointIndex2];
			const SpatialVector normalCalc=
					unitSafe(cross(p1-p0,p2-p0));

			const SpatialVector& n0=normal[pointIndex0];
			const SpatialVector& n1=normal[pointIndex1];
			const SpatialVector& n2=normal[pointIndex2];
			const SpatialVector normalAve=
					unitSafe(n0+n1+n2);

			const Real normalDot=dot(normalCalc,normalAve);
			if(normalDot>Real(0))
			{
				orientation = HdTokens->rightHanded;
			}
		}

		VtVec3fArray normalArray;
		if(normals)
		{
			normalArray=_BuildArray(normals,pointCount);

			delete[] normals;
			normals=NULL;
		}

		if(spBuffer->m_nodeName.empty())
		{
			spBuffer->m_nodeName=
					spBuffer->name().substitute(":","_").substitute(".","_");
		}

		const SdfPath id(("/"+spBuffer->m_nodeName).c_str());
		const GfMatrix4f transform(1);
		const bool guide = false;
		const PxOsdSubdivTags subdivTags;
		const TfToken scheme = allTriangles?
				PxOsdOpenSubdivTokens->loop:
				PxOsdOpenSubdivTokens->catmark;
		const bool doubleSided=false;

		rDelegate.Remove(id);
		m_bufferMap.erase(spBuffer->m_nodeName);

//		feLog("AddMesh %p \"%s\"\n",
//				spBuffer.raw(),spBuffer->m_nodeName.c_str());

		SdfPath instancerId;
		rDelegate.AddMesh(id,transform,
				_BuildArray(points, pointCount),
				_BuildArray(numVerts, faceCount),
				_BuildArray(verts, vertCount),
				normalArray,
				subdivTags,vtColor,interpolation,
				guide,instancerId,scheme,orientation,doubleSided);

		//* NOTE choices: hull, smoothHull, wire, wireOnSurf,
		//*		refined, refinedWire, refinedWireOnSurf
		rDelegate.SetReprName(id,
				drawStyle()==DrawMode::e_wireframe?
				(refined? HdTokens->refinedWire: HdTokens->wire):
				(drawStyle()==DrawMode::e_edgedSolid?
				(refined? HdTokens->refinedWireOnSurf: HdTokens->wireOnSurf):
				(refined? HdTokens->refined: HdTokens->smoothHull)));

		rDelegate.SetRefineLevel(id,refinement);

		spBuffer->m_current=TRUE;

#if FE_DHY_DEBUG
		feLog("DrawHydra::drawPolys new vertices %d\n  %s\n",
				vertices,c_print(drawMode()));
#endif

		delete[] verts;
		verts=NULL;

		delete[] numVerts;
		numVerts=NULL;

		delete[] points;
		points=NULL;

	}

	spBuffer->m_recent=1;

	m_bufferMap[spBuffer->m_nodeName]=spBuffer;

	updateNodes();
}

void DrawHydra::updateNodes(void)
{
	if(!m_pHydraDriver)
	{
		return;
	}

	if(!m_drawn)
	{
		HydraDelegate& rDelegate=m_pHydraDriver->GetDelegate();

//		feLog("---- PRUNING ----\n");

		//* prune old nodes in usd tree
		for(std::map<String, sp<Buffer> >::iterator it=m_bufferMap.begin();
				it!=m_bufferMap.end();)
		{
			String nodeName=it->first;
			sp<Buffer> spBuffer2=it->second;
			it++;

			if(spBuffer2.isNull())
			{
				feLog("DrawHydra::updateNodes invalid buffer for \"%s\"\n",
						nodeName.c_str());
				m_bufferMap.erase(nodeName);
				continue;
			}

			if(spBuffer2->m_recent<1)
			{
#if FE_DHY_DEBUG
				feLog("DrawHydra::updateNodes prune\n  \"%s\"\n",
						spBuffer2->m_nodeName.c_str());
#endif

				m_bufferMap.erase(nodeName);

				const SdfPath id(("/"+spBuffer2->m_nodeName).c_str());
				rDelegate.Remove(id);
				spBuffer2->releaseAll();
				continue;
			}

			spBuffer2->m_recent--;
		}

		const I32 newRefineLevel=drawMode()->refinement();
		if(m_refineLevel!=newRefineLevel)
		{
			//* TODO per mesh

			m_refineLevel=newRefineLevel;
//			feLog("SetRefineLevel global %d\n",m_refineLevel);
			rDelegate.SetRefineLevel(m_refineLevel);
		}

//		feLog("---- REDRAW ----\n");

		updateLighting();

		//* HACK wireframe line width
		glPushAttrib(GL_LINE_BIT);
		glLineWidth(drawMode()->lineWidth());

		m_pHydraDriver->Draw(m_engine);

		glPopAttrib();

		m_drawn=TRUE;
	}
}

} // namespace
