/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "vegetation/vegetation.pmh"

#define	FE_LBO_DEBUG	FALSE

using namespace fe;

LimberOp::LimberOp(void)
{
}

LimberOp::~LimberOp(void)
{
}

void LimberOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<Real>("maxDistance")=1.0;
	catalog<Real>("maxDistance","high")=10.0;
	catalog<Real>("maxDistance","max")=1e6;
	catalog<String>("maxDistance","label")="Max Distance";
	catalog<String>("maxDistance","hint")=
			"The further each root looks for an attachment point.";

	catalog< sp<Component> >("Input Surface");
}

void LimberOp::handle(Record& a_rSignal)
{
#if FE_LBO_DEBUG
	feLog("LimberOp::handle\n");
#endif

	const Real maxDistance=catalog<Real>("maxDistance");

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceI> spInput;
	if(!access(spInput,spInputAccessible)) return;

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<SurfaceAccessorI> spOutputVertices;
	if(!access(spOutputVertices,spOutputAccessible,
			e_primitive,e_vertices)) return;

	sp<SurfaceAccessorI> spOutputAttached;
	if(!access(spOutputAttached,spOutputAccessible,
			e_primitive,"attached")) return;

	const I32 pointCount=spInputAccessible->count(SurfaceAccessibleI::e_point);
	Array<I32> isRoot(pointCount);
	Array<I32> mapping(pointCount);
	for(I32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		isRoot[pointIndex]=FALSE;
		mapping[pointIndex]= -1;
	}

	const I32 primitiveCount=spOutputVertices->count();
	Array<I32> parentPrimitive(primitiveCount);
	for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		parentPrimitive[primitiveIndex]= -1;

		const I32 subCount=spOutputVertices->subCount(primitiveIndex);
		if(subCount)
		{
			const I32 rootIndex=spOutputVertices->integer(primitiveIndex,0);
			isRoot[rootIndex]=TRUE;
		}
	}

	for(I32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const I32 subCount=spOutputVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		const SpatialVector rootPoint=
				spOutputVertices->spatialVector(primitiveIndex,0);
		const I32 hitLimit=8;
		sp<PartitionI> spPartition(NULL);

		Array< sp<SurfaceI::ImpactI> > impactArray=spInput->nearestPoints(
				rootPoint,maxDistance,hitLimit,spPartition);

		const I32 impactCount=impactArray.size();

#if FE_LBO_DEBUG
		const I32 pointIndex=
				spOutputVertices->integer(primitiveIndex,0);

		feLog("prim %d/%d pt %d at %s hits %d\n",
				primitiveIndex,primitiveCount,pointIndex,
				c_print(rootPoint),impactCount);
#endif

		spOutputAttached->set(primitiveIndex,0);

		for(I32 impactIndex=0;impactIndex<impactCount;impactIndex++)
		{
			sp<SurfaceI::ImpactI> spImpact=impactArray[impactIndex];
			const I32 impactPrimitiveIndex=spImpact->primitiveIndex();
			if(impactPrimitiveIndex==primitiveIndex)
			{
				continue;
			}

			BWORD cycle=FALSE;
			I32 checkIndex=impactPrimitiveIndex;
			while(TRUE)
			{
				const I32 attachIndex=parentPrimitive[checkIndex];
				if(attachIndex<0)
				{
					break;
				}
				if(attachIndex==primitiveIndex)
				{
					cycle=TRUE;
					break;
				}
				checkIndex=attachIndex;
			}
			if(cycle)
			{
#if FE_LBO_DEBUG
				feLog("  pr %d cycles\n",impactPrimitiveIndex);
#endif
				continue;
			}

			sp<SurfaceCurves::Impact> spCurveImpact=spImpact;
			if(spCurveImpact.isValid())
			{
				const I32 pointIndex0=spCurveImpact->pointIndex0();
				const I32 pointIndex1=spCurveImpact->pointIndex1();

				const SpatialVector point0=spCurveImpact->vertex0();
				const SpatialVector point1=spCurveImpact->vertex1();

				const SpatialVector arm=unitSafe(point1-point0);
				const SpatialVector toRoot=unitSafe(rootPoint-point0);
				const Real rootDot=dot(arm,toRoot);

#if FE_LBO_DEBUG
				const SpatialBary bary=spImpact->barycenter();
				feLog("  pr %d pt %d,%d along %.6G dot %.6G\n",
						impactPrimitiveIndex,
						pointIndex0,pointIndex1,bary[0],rootDot);
#endif

				if(rootDot<0.0)
				{
					continue;
				}

				const Real mag20=magnitudeSquared(point0-rootPoint);
				const Real mag21=magnitudeSquared(point1-rootPoint);

				I32 hitIndex=(mag20<mag21 && !isRoot[pointIndex0])?
						pointIndex0: pointIndex1;
				const I32 mapIndex=mapping[hitIndex];
				if(mapIndex>=0)
				{
					hitIndex=mapIndex;
				}

				spOutputVertices->set(primitiveIndex,0,hitIndex);

				const I32 rootIndex=
						spOutputVertices->integer(primitiveIndex,0);
				mapping[rootIndex]=hitIndex;

				spOutputAttached->set(primitiveIndex,1);
				parentPrimitive[primitiveIndex]=impactPrimitiveIndex;

				break;
			}
		}
	}

#if FE_LBO_DEBUG
	feLog("LimberOp::handle done\n");
#endif
}
