/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "vegetation/vegetation.pmh"

#define FE_TOP_SIGNAL_DEBUG		FALSE

using namespace fe;

TreeOp::TreeOp(void):
	m_lastFrame(1e6),
	m_lastPointCount(0),
	m_lastPrimitiveCount(0)
{
}

TreeOp::~TreeOp(void)
{
#if FE_TOP_SIGNAL_DEBUG
	Peeker peeker;
	if(m_spSignalerI.isValid())
	{
		m_spSignalerI->peek(peeker);
	}

	const String peek=peeker.str();
	if(!peek.empty())
	{
		feLog("\n~TreeOp\n%s\n",peek.c_str());
	}
#endif
}

void TreeOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* page Config

	catalog<SpatialVector>("TreeLocation");
	catalog<String>("TreeLocation","label")="Tree Location";
	catalog<String>("TreeLocation","page")="Config";
	catalog<String>("TreeLocation","hint")=
			"Displacement of trunk base from origin.";

	catalog<SpatialVector>("gravity")=SpatialVector(0.0,-9.8,0.0);
	catalog<String>("gravity","page")="Config";
	catalog<String>("gravity","label")="Gravity";

	catalog<SpatialVector>("WindDirection");
	catalog<String>("WindDirection","page")="Config";
	catalog<String>("WindDirection","label")="Uniform Wind";

	catalog<bool>("YUp")=true;
	catalog<String>("YUp","label")="Y Is Up";
	catalog<String>("YUp","page")="Config";
	catalog<String>("YUp","hint")=
			"Indicate that the Y axis points up instead of the Z axis.";

	//* page Generate

	catalog<String>("TreeName")="Aspen_Quaking";
	catalog<String>("TreeName","label")="Tree Name";
	catalog<String>("TreeName","page")="Generate";
	catalog<String>("TreeName","hint")="Tree growth configuration"
			" (ignored when input geometry is provided).";

	catalog<String>("stickForm")="none";
	catalog<String>("stickForm","label")="Branch Form";
	catalog<String>("stickForm","choice:0")="none";
	catalog<String>("stickForm","label:0")="None";
	catalog<String>("stickForm","choice:1")="line";
	catalog<String>("stickForm","label:1")="Line";
	catalog<String>("stickForm","choice:2")="cylinder";
	catalog<String>("stickForm","label:2")="Cylinder";
	catalog<String>("stickForm","choice:2")="curve";
	catalog<String>("stickForm","label:2")="Curve";
	catalog<String>("stickForm","page")="Generate";
	catalog<String>("stickForm","hint")=
			"Geometry used to output branch segments.";

	catalog<String>("leafForm")="none";
	catalog<String>("leafForm","label")="Leaf Form";
	catalog<String>("leafForm","choice:0")="none";
	catalog<String>("leafForm","label:0")="None";
	catalog<String>("leafForm","choice:1")="point";
	catalog<String>("leafForm","label:1")="Point";
	catalog<String>("leafForm","choice:2")="circle";
	catalog<String>("leafForm","label:2")="circle";
	catalog<String>("leafForm","page")="Generate";
	catalog<String>("leafForm","hint")=
			"Geometry used to output each leaf.";

	//* page Input

	catalog<String>("velocityAttr")="velocity";
	catalog<String>("velocityAttr","label")="Velocity Attribute";
	catalog<String>("velocityAttr","page")="Input";

	catalog<Real>("rigidity")=1.0;
	catalog<Real>("rigidity","high")=1.0;
	catalog<Real>("rigidity","max")=1e3;
	catalog<String>("rigidity","label")="Rigidity";
	catalog<String>("rigidity","page")="Input";

	catalog<Real>("damping")=0.0;
	catalog<Real>("damping","high")=1.0;
	catalog<Real>("damping","max")=1e3;
	catalog<String>("damping","label")="Damping";
	catalog<String>("damping","page")="Input";

	catalog<Real>("reactivity")=0.1;
	catalog<Real>("reactivity","high")=1.0;
	catalog<Real>("reactivity","max")=10.0;
	catalog<String>("reactivity","label")="Reactivity";
	catalog<String>("reactivity","page")="Input";

	catalog<Real>("threshold")=0.1;
	catalog<Real>("threshold","high")=10.0;
	catalog<Real>("threshold","max")=1e6;
	catalog<String>("threshold","label")="Threshold";
	catalog<String>("threshold","page")="Input";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

	catalog< sp<Component> >("Collider Surface");
	catalog<bool>("Collider Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","temporal")=true;

	//* HACK TBB isn't good with free-form threading
//	Mutex::confirm("fexBoostThread1_56");

	m_spSignalerI=registry()->create("SignalerI");
	sp<HandlerI> spSimPlant=registry()->create("*.SimPlant");
	sp<HandlerI> spDrawPlant=registry()->create("*.DrawPlant");
	if(!m_spSignalerI.isValid() || !spSimPlant.isValid() ||
			!spDrawPlant.isValid())
	{
		feX("TreeOp::initialize", "can't create components");
	}
	m_modifierArray.push_back(spSimPlant);
	m_modifierArray.push_back(spDrawPlant);

	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","PlantScope");
	FEASSERT(m_spScope.isValid());

	m_forestRV.bind(m_spScope);
	m_forestRV.createRecord();
	set(m_forestRV.effectorVelocity());

	m_plantRV.bind(m_spScope);

	m_spSignalerI->insert(spSimPlant,m_forestRV.layout());
	m_spSignalerI->insert(spDrawPlant,m_forestRV.layout());
}

SpatialVector TreeOp::convertY(SpatialVector a_vector)
{
	return m_yUp?
			SpatialVector(a_vector[0],-a_vector[2],a_vector[1]): a_vector;
}

SpatialVector TreeOp::unconvertY(SpatialVector a_vector)
{
	return m_yUp?
			SpatialVector(a_vector[0],a_vector[2],-a_vector[1]): a_vector;
}

BWORD TreeOp::rebuild(sp<SurfaceAccessibleI>& a_rspInputAccessible)
{
	feLog("TreeOp::rebuild\n");

	setName("TreeOp (skeleton)");

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,a_rspInputAccessible,e_primitive,e_vertices))
	{
		return FALSE;
	}

	sp<SurfaceAccessorI> spInputPoints;
	if(!access(spInputPoints,a_rspInputAccessible,e_point,e_generic))
	{
		return FALSE;
	}
	const U32 pointCount=spInputPoints->count();
	const U32 primitiveCount=spInputVertices->count();

	sp<SurfaceAccessorI> spInputRadius;
	access(spInputRadius,a_rspInputAccessible,e_point,"radius",e_quiet);

	sp<SurfaceAccessorI> spInputAttached;
	access(spInputAttached,a_rspInputAccessible,e_primitive,"attached",e_quiet);

	const Real rigidity=catalog<Real>("rigidity");
	const Real damping=catalog<Real>("damping");

	m_forestRV.plants()->clear();

	std::set<U32> pointSet;
	std::set<U32> primitiveSet;
	I32 plantCount=0;
	BWORD allowOneAttached=FALSE;

	while(TRUE)
	{
//		feLog("**** new Tree %d ****\n",plantCount);
		m_plantPointMap.resize(++plantCount);
		m_plantOffset.resize(plantCount);
		std::map<U32,U32>& rPointMap=m_plantPointMap[plantCount-1];

		Record plant=m_plantRV.createRecord();
		m_plantRV.plantName()="<skeleton>";
		m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));

		sp<PlantModelI> spPlantModelI=registry()->create("PlantModelI.Tree");
		if(spPlantModelI.isNull())
		{
			feLog("TreeOp::rebuild could not create Tree\n");
			break;
		}

		m_plantRV.plantModel()=spPlantModelI;

		spPlantModelI->setRigidity(rigidity);
		spPlantModelI->setDamping(damping);

		I32 primitivesAdded=0;
		I32 primitivesSkipped=0;
		I32 nextPointIndex=0;

		while(TRUE)
		{
			I32 primitivesJustAdded=0;

			for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
					primitiveIndex++)
			{
				if(primitiveSet.find(primitiveIndex)!=primitiveSet.end())
				{
					//* already added
					continue;
				}

				const U32 subCount=spInputVertices->subCount(primitiveIndex);
				if(subCount<2)
				{
					continue;
				}

				U32 fromIndex=spInputVertices->integer(primitiveIndex,0);
				FEASSERT(fromIndex<pointCount);

				I32 plantFromIndex=nextPointIndex;

				if(pointSet.find(fromIndex)!=pointSet.end())
				{
					plantFromIndex=rPointMap[fromIndex];

					//* HACK only one root segment
					if(!plantFromIndex)
					{
						plantFromIndex=1;
					}
				}
				else
				{
					if(primitivesAdded)
					{
//						feLog("skipping new root prim %d from %d\n",
//								primitiveIndex,fromIndex);
						primitivesSkipped++;
						continue;
					}

#if FALSE
					const U32 toIndex=
							spInputVertices->integer(primitiveIndex,1);
					if(toIndex!=fromIndex+1)
					{
						feLog("skipping deferred branch %d from %d to %d\n",
								primitiveIndex,fromIndex,toIndex);
						primitivesSkipped++;
						continue;
					}
#else
					if(spInputAttached.isValid() &&
							spInputAttached->integer(primitiveIndex))
					{
						if(allowOneAttached)
						{
							feLog("allow attached prim %d\n",primitiveIndex);
							allowOneAttached=FALSE;
						}
						else
						{
//							feLog("skipping attached prim %d\n",primitiveIndex);
							primitivesSkipped++;
							continue;
						}
					}
#endif

					rPointMap[fromIndex]=nextPointIndex++;
					pointSet.insert(fromIndex);
				}

				SpatialVector fromPosition=convertY(
						spInputVertices->spatialVector(primitiveIndex,0));

				if(!primitivesAdded)
				{
					m_plantOffset[plantCount-1]=fromPosition;
				}

				primitiveSet.insert(primitiveIndex);
				primitivesAdded++;
				primitivesJustAdded++;

				Real curveLength(0);
				Real segmentLength[subCount-1];
				SpatialVector basePosition=fromPosition;
				for(U32 subIndex=1;subIndex<subCount;subIndex++)
				{
					const SpatialVector tipPosition=convertY(
							spInputVertices->spatialVector(
							primitiveIndex,subIndex));

					Real& rSegmentLength=segmentLength[subIndex-1];
					rSegmentLength=magnitude(tipPosition-basePosition);
					curveLength+=rSegmentLength;

					basePosition=tipPosition;

//~					feLog("  seg %d/%d length %.6G\n",
//~							subIndex-1,subCount-1,rSegmentLength);
				}

//~				feLog("curve %d from %d map %d curveLength %.6G\n",
//~						primitiveIndex,fromIndex,plantFromIndex,
//~						curveLength);

				Real remainingLength=curveLength;

				for(U32 subIndex=1;subIndex<subCount;subIndex++)
				{
					const U32 toIndex=
							spInputVertices->integer(primitiveIndex,subIndex);
					FEASSERT(toIndex<pointCount);

					const SpatialVector toPosition=convertY(
							spInputVertices->spatialVector(
							primitiveIndex,subIndex));
					if(pointSet.find(toIndex)!=pointSet.end())
					{
						feLog("TreeOp::rebuild redundant segment"
								" from point %d to %d\n",
								fromIndex,toIndex);
						fromIndex=toIndex;
						plantFromIndex=rPointMap[toIndex];
						fromPosition=toPosition;
						continue;
					}

					const SpatialVector span=toPosition-fromPosition;
					const Real mag2=magnitudeSquared(span);
					if(mag2<1e-6)
					{
						feLog("TreeOp::rebuild short segment"
								" from point %d to %d length %.6G\n",
								fromIndex,toIndex,sqrt(mag2));

						//* TODO should move redundant point in replace()
						continue;
					}

					I32 plantToIndex=nextPointIndex;
					rPointMap[toIndex]=nextPointIndex++;

					const U32 level=(primitiveIndex!=0);

					Real radius=1.0;
					if(spInputRadius.isValid())
					{
						radius=spInputRadius->real(toIndex);
					}
					else
					{
#if FALSE
						//* HACK radius by distance from root
						Real radiusFade=magnitude(toPosition)/30.0;
						if(radiusFade>1.0)
						{
							radiusFade=1.0;
						}
						const Real radius=0.10-0.09*radiusFade;
#else
						radius=0.01*remainingLength;
						remainingLength-=segmentLength[subIndex-1];
#endif
					}

//					feLog("  seg %d/%d pt %d radius %.6G from %d to %d\n",
//							subIndex-1,subCount-1,toIndex,radius,
//							plantFromIndex,plantToIndex);

					spPlantModelI->addBranch(plantFromIndex,plantToIndex,
							level,radius,span);

					fromIndex=toIndex;
					plantFromIndex=plantToIndex;
					fromPosition=toPosition;
					pointSet.insert(toIndex);
				}
			}

			if(!primitivesJustAdded)
			{
				break;
			}
		}

		if(primitivesAdded)
		{
			spPlantModelI->prepare();

			m_forestRV.plants()->add(plant);
		}

		m_plantRV.unbind();

//		feLog("primitivesAdded %d primitivesSkipped %d\n",
//				primitivesAdded,primitivesSkipped);

		if(!primitivesSkipped)
		{
			break;
		}
		else if(!primitivesAdded)
		{
			allowOneAttached=TRUE;
		}
	}

	return TRUE;
}

BWORD TreeOp::updateWind(sp<SurfaceAccessibleI>& a_rspInputAccessible)
{
	const String velocityAttr=catalog<String>("velocityAttr");
	if(velocityAttr.empty())
	{
		return FALSE;
	}

	sp<SurfaceAccessorI> spInputVelocity;
	if(!access(spInputVelocity,a_rspInputAccessible,
			e_point,velocityAttr,e_quiet))
	{
		return FALSE;
	}

	sp<RecordGroup> spRG=m_forestRV.plants();
	RecordGroup::iterator it=spRG->begin();
	if(it==spRG->end())
	{
		return FALSE;
	}
	sp<RecordArray> spRA=*it;

	const U32 plantCount=spRA->length();
	if(!plantCount)
	{
		return FALSE;
	}

	for(U32 plantIndex=0;plantIndex<plantCount;plantIndex++)
	{
		Record plant=spRA->getRecord(plantIndex);
		if(!plant.isValid())
		{
			continue;
		}
		m_plantRV.bind(plant);

		sp<PlantModelI> spPlantModelI=m_plantRV.plantModel();

		m_plantRV.unbind();

		if(spPlantModelI.isNull())
		{
			feLog("TreeOp::updateWind PlantModelI is NULL\n");
			continue;
		}

		const U32 numSticks=spPlantModelI->numSticks();

		std::map<U32,U32>& rPointMap=m_plantPointMap[plantIndex];
		for(std::map<U32,U32>::iterator it=rPointMap.begin();
				it!=rPointMap.end();it++)
		{
			const U32 pointIndex=it->first;
			const U32 plantPointIndex=it->second;

			const SpatialVector velocity=
					spInputVelocity->spatialVector(pointIndex);

			if(plantPointIndex<=numSticks)
			{
				PlantModelI::Stick* pStick=plantPointIndex?
						spPlantModelI->stick(plantPointIndex-1):
						spPlantModelI->stick(0);
				if(pStick)
				{
					pStick->setWindVelocity(velocity);
				}
			}

//			feLog("plant %d/%d map %d %d vel %s\n",
//					plantIndex,plantCount,pointIndex,plantPointIndex,
//					c_print(velocity));
		}
	}

	return TRUE;
}

void TreeOp::replace(sp<SurfaceAccessibleI>& a_rspOutputAccessible)
{
	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,a_rspOutputAccessible,e_point,e_position))
	{
		return;
	}

	sp<SurfaceAccessorI> spOutputRadius;
	if(!access(spOutputRadius,a_rspOutputAccessible,e_point,"radius"))
	{
		return;
	}

	sp<RecordGroup> spRG=m_forestRV.plants();
	RecordGroup::iterator it=spRG->begin();
	if(it==spRG->end())
	{
		return;
	}
	sp<RecordArray> spRA=*it;

	const U32 plantCount=spRA->length();
	if(!plantCount)
	{
		return;
	}

	for(U32 plantIndex=0;plantIndex<plantCount;plantIndex++)
	{
		Record plant=spRA->getRecord(plantIndex);
		if(!plant.isValid())
		{
			continue;
		}
		m_plantRV.bind(plant);

		sp<PlantModelI> spPlantModelI=m_plantRV.plantModel();
		if(spPlantModelI.isNull())
		{
			feLog("TreeOp::replace %d/%d \"%s\" PlantModelI is NULL\n",
					plantIndex,plantCount,
					m_plantRV.plantName().c_str());
			m_plantRV.unbind();
			continue;
		}

//		const SpatialVector root=
				unconvertY(m_plantRV.location());

		const U32 numSticks=spPlantModelI->numSticks();

		const SpatialVector offset=m_plantOffset[plantIndex];

		std::map<U32,U32>& rPointMap=m_plantPointMap[plantIndex];
		for(std::map<U32,U32>::iterator it=rPointMap.begin();
				it!=rPointMap.end();it++)
		{
			const U32 plantPointIndex=it->second;

#if TRUE
			if(plantPointIndex<=numSticks)
			{
				const PlantModelI::Stick* pStick=plantPointIndex?
						spPlantModelI->stick(plantPointIndex-1):
						spPlantModelI->stick(0);
				if(pStick)
				{
					const U32 pointIndex=it->first;
					const SpatialVector& base=pStick->base();
					const SpatialVector& span=pStick->span();
					const SpatialVector point=unconvertY(plantPointIndex?
							offset+base+span: offset+base);

					spOutputPoint->set(pointIndex,point);
					spOutputRadius->set(pointIndex,pStick->radius1());
				}
			}
#else
			const U32 pointIndex=it->first;
			Real radius=0.0;

//			SpatialVector point=
//					root+offset+spOutputPoint->spatialVector(pointIndex);
			SpatialVector point(0.0,0.0,0.0);

			if(plantPointIndex<=numSticks)
			{
				const PlantModelI::Stick* pStick=plantPointIndex?
						spPlantModelI->stick(plantPointIndex-1):
						spPlantModelI->stick(0);
				if(pStick)
				{
					const SpatialVector base=pStick->base();
					const SpatialVector span=pStick->span();
					point=unconvertY(plantPointIndex?
							offset+base+span: offset+base);
					radius=pStick->radius1();
				}
			}

//			feLog("plant %d/%d map %d %d point %s\n",
//					plantIndex,plantCount,pointIndex,plantPointIndex,
//					c_print(point));

			spOutputPoint->set(pointIndex,point);
			spOutputRadius->set(pointIndex,radius);
#endif
		}

		m_plantRV.unbind();
	}
}

void TreeOp::regenerate(void)
{
	feLog("TreeOp::regenerate \"%s\"\n",m_treeName.c_str());

	setName(m_treeName);

	m_forestRV.plants()->clear();

	const String path=
			registry()->master()->catalog()->catalog<String>("path:media")+
			"/vegetation/"+m_treeName+".rg";

	sp<RecordGroup>	spRG=RecordView::loadRecordGroup(m_spScope,path);
	if(spRG.isNull())
	{
		feLog("TreeOp::regenerate failed to load:\n  \"%s\"\n",path.c_str());
		return;
	}

	RecordGroup::iterator it=spRG->begin();
	sp<RecordArray>& rspRA= *it;
	if(m_plantRV.plantName.check(rspRA))
	{
		m_forestRV.plants()->add(spRG);
	}
	else
	{
		Record plant=m_plantRV.createRecord();
		m_plantRV.plantName()=m_treeName;
		m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));
		m_forestRV.plants()->add(plant);
	}

	m_plantRV.unbind();
}

void TreeOp::handle(Record& a_rSignal)
{
	const Real frame=currentFrame(a_rSignal);
	const Real firstFrame=startFrame(a_rSignal);
	m_time=currentTime(a_rSignal);

	m_yUp=catalog<bool>("YUp");

//	const BWORD outputReplaced=catalog<bool>("Output Surface","replaced");
//	feLog("TreeOp::handle start %.6G frame %.6G time %.6G\n",
//			start,frame,m_time);

	sp<SurfaceAccessibleI> spInputAccessible;
	access(spInputAccessible,"Input Surface",e_quiet);
	if(spInputAccessible.isValid())
	{
		const I32 primitiveCount=
				spInputAccessible->count(SurfaceAccessibleI::e_primitive);
		const I32 pointCount=
				spInputAccessible->count(SurfaceAccessibleI::e_point);

//		feLog("TreeOp::handle prims %d vs %d pts %d vs %d\n",
//				primitiveCount,m_lastPrimitiveCount,
//				pointCount,m_lastPointCount);

		m_treeName="";
		if(frame==firstFrame || frame<m_lastFrame ||
				primitiveCount!=m_lastPrimitiveCount ||
				pointCount!=m_lastPointCount)
		{
			if(!rebuild(spInputAccessible))
			{
				return;
			}
		}

		updateWind(spInputAccessible);

		catalog<String>("summary")="using input";

		m_lastPrimitiveCount=primitiveCount;
		m_lastPointCount=pointCount;
	}
	else
	{
		const String treeName=catalog<String>("TreeName");
		if(m_treeName!=treeName)
		{
			feLog("TreeOp::handle treeName changed \"%s\" -> \"%s\"\n",
					m_treeName.c_str(),treeName.c_str());
			m_treeName=treeName;
			regenerate();
		}

		catalog<String>("summary")=m_treeName;
	}

	m_lastFrame=frame;

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	m_forestRV.upAxis()=catalog<bool>("YUp")? 1: 2;

#if TRUE
	//* HACK always draw (for Arnold tests)
	m_forestRV.drawI()=spDrawI;
#else
	m_forestRV.drawI()=spInputAccessible.isValid()? sp<DrawI>(NULL): spDrawI;
#endif

	m_forestRV.stickForm()=catalog<String>("stickForm");
	m_forestRV.leafForm()=catalog<String>("leafForm");
	m_forestRV.deltaT()=(1.0f/24.0f);
	m_forestRV.gravity()=
			convertY(catalog<SpatialVector>("gravity"));
	m_forestRV.uniformVelocity()=
			convertY(catalog<SpatialVector>("WindDirection"));

	sp<SurfaceI> spCollider;
	access(spCollider,"Collider Surface",e_quiet);

#if FALSE
	if(spCollider.isValid())
	{
		sp<SurfaceSphere> spSphere(new SurfaceSphere());
		spSphere->setCenter(convertY(spCollider->center()));
		spSphere->setRadius(spCollider->radius());

		spCollider=spSphere;
	}
#endif

	SpatialTransform colliderTransform;
	setIdentity(colliderTransform);

	if(catalog<bool>("YUp"))
	{
		colliderTransform.column(0)=SpatialVector(1.0, 0.0, 0.0);
		colliderTransform.column(1)=SpatialVector(0.0, 0.0, 1.0);
		colliderTransform.column(2)=SpatialVector(0.0,-1.0, 0.0);
		set(colliderTransform.translation());
	}

	m_forestRV.collider()=spCollider;
	m_forestRV.colliderTransform()=colliderTransform;

	sp<RecordGroup>& rspPlants=m_forestRV.plants();
	FEASSERT(rspPlants.isValid());

	//* update location, but only if it is an individual tree
	for(RecordGroup::iterator it=rspPlants->begin();it!=rspPlants->end();it++)
	{
		sp<RecordArray> spRA= *it;

		if(spRA->length()!=1)
		{
			continue;
		}

		if(!m_plantRV.plantModel.check(spRA))
		{
			continue;
		}

		m_plantRV.bind(spRA);
		while(m_plantRV.step())
		{
			m_plantRV.location()=
					convertY(catalog<SpatialVector>("TreeLocation"));
		}
	}

	const Real reactivity=catalog<Real>("reactivity");
	const Real threshold=catalog<Real>("threshold");

	for(RecordGroup::iterator it=rspPlants->begin();it!=rspPlants->end();it++)
	{
		sp<RecordArray> spRA= *it;

		if(!m_plantRV.plantModel.check(spRA))
		{
			continue;
		}

		U32 plantIndex=0;

		m_plantRV.bind(spRA);
		while(m_plantRV.step())
		{
			m_plantRV.threshold()=threshold;
			m_plantRV.reactivity()=reactivity;

			m_plantRV.offset()=plantIndex<m_plantOffset.size()?
					m_plantOffset[plantIndex]: SpatialVector(0.0,0.0,0.0);
			plantIndex++;
		}
	}

	m_plantRV.unbind();

	Record forest=m_forestRV.record();

	m_spSignalerI->signal(forest);

	if(spInputAccessible.isValid())
	{
		sp<SurfaceAccessibleI> spOutputAccessible;
		accessOutput(spOutputAccessible,a_rSignal);

		replace(spOutputAccessible);
	}

//	feLog("TreeOp::handle done\n");
}
